#pragma once

#include <type_traits>

#include <amdis/GridFunctionOperator.hpp>
#include <amdis/LocalBasisCache.hpp>
#include <amdis/common/ValueCategory.hpp>

namespace AMDiS
{
  /**
   * \addtogroup operators
   * @{
   **/

  namespace tag
  {
    struct test {};
  }


  /// zero-order vector-operator \f$ (c\, \psi) \f$
  template <class LocalContext, class GridFct>
  class GridFunctionOperator<tag::test, LocalContext, GridFct>
      : public GridFunctionOperatorBase<GridFunctionOperator<tag::test, LocalContext, GridFct>,
                                        LocalContext, GridFct>
  {
    using Self = GridFunctionOperator;
    using Super = GridFunctionOperatorBase<Self, LocalContext, GridFct>;

    static_assert( Category::Scalar<typename GridFct::Range>, "Expression must be of scalar type." );

  public:
    GridFunctionOperator(tag::test, GridFct const& expr)
      : Super(expr, 0)
    {}

    template <class Context, class Node, class ElementVector>
    void getElementVector(Context const& context,
                          Node const& node,
                          ElementVector& elementVector)
    {
      static_assert(Node::isLeaf, "Operator can be applied to Leaf-Nodes only");

      auto const& quad = this->getQuadratureRule(context.type(), node);
      auto const& localFE = node.finiteElement();
      std::size_t size = localFE.size();

      using LocalBasisType = typename std::decay_t<decltype(localFE)>::Traits::LocalBasisType;
      LocalBasisCache<LocalBasisType> cache(localFE.localBasis());

      auto const& shapeValuesCache = cache.evaluateFunctionAtQp(context,quad);
      for (std::size_t iq = 0; iq < quad.size(); ++iq) {
        // Position of the current quadrature point in the reference element
        decltype(auto) local = context.local(quad[iq].position());

        // The multiplicative factor in the integral transformation formula
        const auto factor = Super::coefficient(local) * context.integrationElement(quad[iq].position())
                                                      * quad[iq].weight();

        auto const& shapeValues = shapeValuesCache[iq];
        for (std::size_t i = 0; i < size; ++i) {
          const auto local_i = node.localIndex(i);
          elementVector[local_i] += factor * shapeValues[i];
        }
      }
    }
  };

  /** @} **/

} // end namespace AMDiS
